#!/usr/bin/env python3


import argparse
import itertools
import json
import matplotlib.pyplot as plt
import numpy as np
import random
import sys
import time
import tqdm 
import gc

from timeit import default_timer as timer



#### STRUCTURE ####

# Each algorithm is taking the following arguments
# - 'files_requested':  a list of file index on the tape that are requested 
# - 'request_numbers': a list containing how many times each file in 'files_requested' is requested
# - 'tape': a list of the all file sizes on the tape
# - 'right': a list of the right ordinate of each file in 'tape'
# !!! NOTE : tape starts by an unrequested file to allow the use of right[x-1] to get the left ordinate of the x-th file !!!

# The output of an algorithm is a list of detours: a couple (a,b) which means that the
# head goes to the left of file a then to the right of file b>=a. a=0 means the leftmost
# file requested. The detours are ordered in chronological order, so by decreasing
# values of a. The leftmost requested file is not included in this list but automatically
# accounted for inside a final additional detour reading all remaining files.



#### COMPUTING THE COST OF AN ALGORITHM ####

# given the detours done by an algorithm, return the sum of completion times
def compute_cost_detours(detours, files_requested, request_numbers, tape, right, penalty_uturn):
    remaining = [0 for i in tape]
    for f in range(len(files_requested)):
        remaining[files_requested[f]] = request_numbers[f]
    last_pos = len(tape)-1 # right of the file indicated
    cur_travelled = 0
    cur_cost = 0
    aux_detours = detours + [(0,len(files_requested)-1)] # add the final detour
    previa = len(files_requested)
    for (ia,ib) in aux_detours:
        assert (ia < previa)
        previa = ia
        
        a = files_requested[ia]
        b = files_requested[ib]
        cur_travelled += right[last_pos]-right[a-1] + penalty_uturn
        last_pos = a-1
        for i,x in enumerate(remaining[a:b+1]):
            if x>0:
                cost = x * (cur_travelled + right[a+i]-right[last_pos])
                cost -= x * (right[-1] - right[a+i]+2*tape[a+i] + penalty_uturn ) # mandatory cost not counted in the DP
                cur_cost += cost
                remaining[a+i]=0
        cur_travelled += 2*(right[b]-right[last_pos]) + penalty_uturn
    return cur_cost



#### COMPUTATION OF THE LOWER BOUND ####

def LB(files_requested, request_numbers, tape, right, penalty_uturn):
    cost = 0
    for i,f in enumerate(files_requested):
        cost += (right[-1] - right[f] + 2*tape[f]+ penalty_uturn) * request_numbers[i]
    return cost










#### ALGORITHMS ####


### Dynamic programming
# T[(ia,iab,nskip)][0] contains the optimal cost charged to remaining files between ia and ib=ia+iab when a detour 
# between files_requested[ia] and >= files_requested[ib] is done with nskip skipped files.
# T[(ia,iab,nskip)][1] contains a best recursive call (c) to follow
# allow to cap the detour sizes in order to improve the running time but becomes suboptimal
def DP(files_requested, request_numbers, tape, right, penalty_uturn, max_detour_length=-1):

    total_requests = sum(request_numbers) 
    nb_files_requested = len(files_requested)
    nl = np.cumsum([0]+request_numbers) # number of requests on the left of each requested file
    length = right[-1]
    
    T = dict()
    
    # recursive function to fill the relevant table cells, returns the optimal cost
    def rec_DP(ia,iab,nskip):        
    
        # the cell has already been computed
        if (ia,iab,nskip) in T:
            return T[(ia,iab,nskip)][0]
        
        assert(iab >= 0)
        
        # base case: a single file is considered
        if iab == 0:
            res = (2*tape[files_requested[ia]])*(nskip+nl[ia])
            T[(ia,iab,nskip)] = (res,-1)
            return res
        
        # main case: test if ib is skipped or all detours between ia and ib
        
        ib = ia + iab # index of the rightmost file considered
        nleft = nskip + nl[ia]
        rb = right[files_requested[ib]]
        lb = right[files_requested[ib]-1]
        rb_l = right[files_requested[ib-1]]
        
        # 1st possibility: b is skipped
        best = rec_DP(ia,iab-1, nskip+request_numbers[ib]) + 2*(rb-rb_l)*(nleft) + 2*(lb-rb_l)*request_numbers[ib]
        res_rec = iab+1 # value easy to recognize, makes no sense for the other case
        
        # try a detour (ic,ib) for all ic > ia and <= ib, respecting ib-ic < max_detour_length
        leftmost_detour = 0 if max_detour_length == -1 else max(0,iab-max_detour_length)
        for iac in range(iab,leftmost_detour,-1): # reverse order as low ic are expected to be expensive so might not need the second recursive call
            cur = 2*(rb-right[files_requested[ia+iac-1]])*nleft+ 2*penalty_uturn*(nskip+nl[ia+iac])
            if (cur >= best):
                continue
            cur += rec_DP(ia+iac,iab-iac,nskip) 
            if (cur >= best):
                continue
            cur += rec_DP(ia,iac-1,nskip) 
            if cur < best:
                best = cur
                res_rec = iac # saves which recursive call is the best choice
    
        # save the cost and best recursive call
        T[(ia,iab,nskip)] = (best,res_rec)
        
        
        return best
        
    cost = rec_DP(0,nb_files_requested-1,0)
    
    # go through the table to extract the detours corresponding to the optimal cost
    def build_detours(ia,iab,nskip):
        if iab == 0:
            return []
        iac = T[(ia,iab,nskip)][1]
        if iac > iab:
            return build_detours(ia,iab-1, nskip+request_numbers[iab+ia]) # b was skipped
        return build_detours(ia+iac,iab-iac,nskip) + [(ia+iac,ia+iab)] + build_detours(ia,iac-1,nskip)
        
    detours = build_detours(0,nb_files_requested-1,0)
    return detours


### log_DP_x but limits the detour maximum size to be x*logarithmic in the number of files requested
def log_DP_five(files_requested, request_numbers, tape, right, penalty_uturn):
    cap = int(np.log(len(files_requested)))*5
    return DP(files_requested, request_numbers, tape, right, penalty_uturn, cap)
    
def log_DP_one(files_requested, request_numbers, tape, right, penalty_uturn):
    cap = int(np.log(len(files_requested)))
    return DP(files_requested, request_numbers, tape, right, penalty_uturn, cap)
    


### Algorithm going directly to the leftmost requested file, can be arbitrarily bad
def No_detour(files_requested, request_numbers, tape, right, penalty_uturn):
    return []
    
    
### Algorithm doing a detour to read each file separately
def GS(files_requested, request_numbers, tape, right, penalty_uturn):
    detours = [ (a,a) for a in range(1,len(files_requested))]
    detours.reverse()
    return detours


### GS but filters out detrimental detours
def FGS(files_requested, request_numbers, tape, right, penalty_uturn):
    detours = [ a for a in range(1,len(files_requested))]
    continuer = True
    
    # continue until we do a full pass without removing a detour
    while(continuer):
        continuer = False
        
        # go through the requested files in increasing order (so first, f=0)
        
        left_detours_size = 0 # = sum_{f'<f and (f,f) is a detour} tape[file[f']]
        left_requests = 0 # sum_{f'<f} request_numbers[file[f']]
        right_skipped = sum(request_numbers) - sum([request_numbers[f] for f in detours])  
        # sum_{f'>f and f' skipped} request_numbers[file[f']] 
        left_detours = [] # detours < f
        right_detours = detours # detours >= f
        
        # main loop: we go through the requested files and test if it is beneficial to remove a detour
        for f in range(len(files_requested)):
            
            if len(right_detours) == 0 or f < right_detours[0]:
                right_skipped -= request_numbers[f] # f was skipped, so just update right_skipped
            else:
            
                lhs = request_numbers[f]*(right[files_requested[f]-1]-right[files_requested[0]-1]+ left_detours_size)
                rhs = (tape[files_requested[f]] + penalty_uturn )* ( left_requests+right_skipped )

                assert (f == right_detours.pop(0))            
                
                if (lhs >= rhs): 
                    # condition to keep the detour (f,f): save it and update left_detours_size
                    left_detours.append(f)
                    left_detours_size += tape[files_requested[f]] + penalty_uturn
                else:
                    # we remove a detour so will do another pass
                    continuer = True
        
            # update left_requests in all cases
            left_requests += request_numbers[f]
        
        # the new detours are the ones that have been saved
        detours = left_detours
    
    detours = [(f,f) for f in detours]
    detours.reverse()
    return detours


### NGS but adds back some non-unit detour if they can be detected as beneficial for totally skipped files
def NFGS(files_requested, request_numbers, tape, right, penalty_uturn, max_detour_length=-1):
    detours = FGS(files_requested, request_numbers, tape, right, penalty_uturn)
    detours.reverse()
    
    
    left_requests = 0
    left_detours = []
    right_detours = detours
    rightest_added_detour = -10
    
    total_detour_left = [0 for _ in files_requested] 
    cum = 0
    detour_index = 0
    if len(detours)>0:
        for i in range(len(total_detour_left)):
            total_detour_left[i] = cum
            if detour_index < len(detours) and i == detours[detour_index][0]:
                cum += tape[files_requested[i]]
                detour_index += 1
    
    # main loop
    for f in range(0,len(files_requested)):
        
        was_a_detour = False

        # remove the (f,f) detour, update total_detour_left
        if len(right_detours) > 0 and f == right_detours[0][0]:
            was_a_detour = True
            right_detours.pop(0)
            sf = tape[files_requested[f]]
            for f2 in range(f+1, len(files_requested)):
                total_detour_left[f2] -= sf
        
        best_delta = 0
        best_g = -1
        
        # inner loop, find the best (f,g) detour if needed
        right_g = len(files_requested)
        if max_detour_length>0:
            right_g = min(right_g, f+max_detour_length)
        for g in range(f, right_g):
            
            # formulas to compute Delta, with b equal to the detour (f,g)
            
            # r(g) - l(f) + Uturn penalty
            sb = right[files_requested[g]] - right[files_requested[f]-1] +  penalty_uturn
            # sum_{x>b, x skipped} x(x). Skipped files are after rightest_added_detour and not in right_detours
            right_skipped = sum([request_numbers[x] for x in range(max(g,rightest_added_detour)+1,len(request_numbers))]) - \
                            sum([request_numbers[x] if x > max(g,rightest_added_detour) else 0 for (x,_) in right_detours])
            
            # first (positive) term of Delta
            lhs = sb*(left_requests + right_skipped)
            
            # sum_{x in (f,g) but currently completely skipped} x(x)*(l(f) + sum_{b' detour with l(b')<f} (delay of the detour b'))
            def rhs_x(x):
                return request_numbers[x] * (right[files_requested[f]-1]-right[files_requested[0]-1] + total_detour_left[f])
            #second (negative) term of Delta
            rhs = sum([rhs_x(x) for x in range(max(f,rightest_added_detour+1),g+1)]) - \
                  sum([rhs_x(x) if (x>rightest_added_detour and x <= g) else 0 for (x,_) in right_detours])

            delta = lhs - rhs
            
            
            # save the best detour candidate
            if delta < best_delta:
                best_delta = delta
                best_g = g
        
        if best_delta == 0 and was_a_detour and rightest_added_detour > f: # additional case where we keep unit detours part of a big detour
            best_g = f
            best_delta = -1
                
        if (best_delta < 0 and f>0): # do not add detours on the leftmost requested file as they are automatically added later
            left_detours.append((f,best_g))
            rightest_added_detour = max(rightest_added_detour,best_g)
            sb = right[files_requested[best_g]] - right[files_requested[f]-1] + penalty_uturn
            for f2 in range(best_g+1, len(files_requested)):
                total_detour_left[f2] += sb
        
        
                    
        left_requests += request_numbers[f]
        
        
    detours = left_detours
    
    detours.reverse()
    return detours




### NFGS but limits the detour maximum size to be 5*logarithmic in the number of files requested
def log_NFGS(files_requested, request_numbers, tape, right, penalty_uturn):
    cap = int(np.log(len(files_requested)))*5
    return NFGS(files_requested, request_numbers, tape, right, penalty_uturn, cap)



