#!/usr/bin/env python3


import numpy as np
import sys
import time
import csv
import os
import gc
from operator import add
from collections import deque
from algorithms import *
from draw import *



def run_all_algos(tape,files_requested,request_numbers, penalty_uturn, outfile, tape_name):

    #list to save algo performance for given tape name
    res=[tape_name]

    print("\n\nNumber of files requested: "+str(len(files_requested)))
    print("Size of tape: "+str(len(tape)-1)+"\n")

    assert(files_requested[0]>0)
    assert(sorted(files_requested) == files_requested)
    right = np.cumsum(tape)

    LB_cost=LB(files_requested,request_numbers,tape,right, penalty_uturn)
    res.append(LB_cost)

    print("LB to add to each algorithm cost: ", LB_cost)
    # run each algorithm on the current input
    for alg in [DP,log_DP_five, log_DP_one, No_detour, GS, FGS, NFGS, log_NFGS]:
        print(alg.__name__)
        start = time.time()
        detours = alg(files_requested,request_numbers,tape, right, penalty_uturn)
        end = time.time()
        cost=compute_cost_detours(detours, files_requested, request_numbers, tape, right, penalty_uturn)
        print("\t" + str(detours))
        print("\t" + str(cost))
        print("\t", int(end-start), " seconds")

        try:
            os.makedirs("draw/"+tape_name+"/")
        except FileExistsError:
            # directory already exists
            pass
        # draw the trajectory of the head for the output sequence
        draw(detours, files_requested,tape, right, alg.__name__ +str(penalty_uturn),'draw/'+tape_name)

        # save results
        res.append(cost)
        res.append(float(end-start))
        
        gc.collect()

    # save the results in output file
    with open(outfile, 'a', encoding='UTF8', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(res)
    f.close()


def main():

    # Value of the u_turn penalty:
    # 0 -> no penalty
    # 14254750000 -> average half segment size over all tape inputs
    # 28509500000 -> average segment size over all tape inputs
    penalty_uturn = 0 

    # open the file with the list of all tapes
    tapes=list()
    with open("../input/list_of_tape.txt") as f:
        tapes = f.read().splitlines() 


    # path to the file to save algorithm performance 
    file="output.csv"
    header = ['tape','LB_cost','perf_DP','time_DP','perf_log_DP_five','time_log_DP_five','perf_log_DP_one','time_log_DP_one','perf_No_detour','time_No_detour','perf_GS','time_GS','perf_FGS','time_FGS','perf_NFGS','time_NFGS','perf_log_NFGS','time_log_NFGS']
    with open(file, 'w', encoding='UTF8', newline='') as f:
        writer = csv.writer(f)
        writer.writerow(header)
    f.close()

    # iterate over each tape in the input file
    for target_tape in tapes:

        print("Tape: ", target_tape)
        input_file = "../input/requests/"+target_tape
        segment_file = "../input/tapes/"+target_tape

        input_data = {}
        segment_data = {}

        # open the file in universal line ending mode 
        with open(input_file, 'r') as infile:
            # read the file as a dictionary for each row ({header : value})
            reader = csv.DictReader(infile)
            for row in reader:
                for header, value in row.items():
                    try:
                        input_data[header].append(value)
                    except KeyError:
                        input_data[header] = deque()
                        input_data[header].append(value)

        # open the file in universal line ending mode 
        with open(segment_file, 'r') as infile:
            # read the file as a dictionary for each row ({header : value})
            reader = csv.DictReader(infile)
            for row in reader:
                for header, value in row.items():
                    try:
                        segment_data[header].append(value)
                    except KeyError:
                        segment_data[header] = deque()
                        segment_data[header].append(value)


        # list of file index requested
        files_requested = input_data['index']
        files_requested=list(map(int, files_requested))


        # how many each file is requested
        request_numbers = input_data['nb_requests']
        request_numbers=list(map(int, request_numbers))

        # list of the file sizes on the tape
        tape = segment_data['segment_size']
        tape=deque(map(int, tape))
        # add unrequested file in position 0 (in order to have access to a file in position -1)    
        tape.appendleft(-1)
        tape=list(tape)



        # get performance of each algorithm for given input tape list of requests
        run_all_algos(tape,files_requested,request_numbers,penalty_uturn, file,target_tape)   
        
            
    return 0

    

if __name__ == '__main__':
  main()
